"""Test extractor."""
# pylint: disable=missing-class-docstring,missing-function-docstring
import re
import unittest

import reticker


class TestExtraction(unittest.TestCase):
    def setUp(self) -> None:
        self.default_ticker_extractor = reticker.TickerExtractor()

    def test_pattern_type(self):
        self.assertIsInstance(self.default_ticker_extractor.pattern, re.Pattern)

    def test_default_config(self):
        text = "Comparing FNGU vs $WEBL vs SOXL- who wins? And what about $cldl vs $Skyu? BTW, will the $w+$Z pair still grow? IMHO, SOXL is king!"
        expected = ["FNGU", "WEBL", "SOXL", "CLDL", "SKYU", "W", "Z"]
        extracted = self.default_ticker_extractor.extract(text)
        self.assertEqual(expected, extracted)

    def test_no_matches(self):
        self.assertEqual(self.default_ticker_extractor.extract("Test text"), [])
        self.assertEqual(self.default_ticker_extractor.extract(""), [])

    def test_deduplicate(self):
        text = "SPY is not QQQ. It is SPY."
        self.assertEqual(reticker.TickerExtractor().extract(text), ["SPY", "QQQ"])
        self.assertEqual(reticker.TickerExtractor(deduplicate=True).extract(text), ["SPY", "QQQ"])
        self.assertEqual(reticker.TickerExtractor(deduplicate=False).extract(text), ["SPY", "QQQ", "SPY"])

    def test_custom_config(self):
        assert "WEBL" not in reticker.BLACKLIST
        assert "TECL" not in reticker.BLACKLIST

        self.assertEqual(reticker.TickerExtractor().extract("WEBL $TECL"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(unprefixed_uppercase=True)).extract("WEBL $TECL"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(unprefixed_uppercase=False)).extract("WEBL $TECL"), ["TECL"])

        self.assertEqual(reticker.TickerExtractor().extract("WEBL $tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig()).extract("WEBL $tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(prefixed_lowercase=True)).extract("WEBL $tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(prefixed_lowercase=False)).extract("WEBL $tecl"), ["WEBL"])

        self.assertEqual(reticker.TickerExtractor().extract("WEBL $Tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig()).extract("WEBL $Tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(prefixed_titlecase=True)).extract("WEBL $Tecl"), ["WEBL", "TECL"])
        self.assertEqual(reticker.TickerExtractor(match_config=reticker.TickerMatchConfig(prefixed_titlecase=False)).extract("WEBL $Tecl"), ["WEBL"])



"""Package config."""
from pathlib import Path
from typing import Final, List, Set

_CONFIG_PATH: Final[Path] = Path(__file__).parent
BLACKLIST_PATHS: Final[List[Path]] = list((_CONFIG_PATH / "blacklist").glob("*.txt"))
BLACKLIST: Set[str] = set(term for path in BLACKLIST_PATHS for term in path.read_text().strip().split("\n"))



"""Package implementation."""
import functools
import re
from typing import List, Optional, Pattern, Set

from . import config


class TickerMatchConfig:
    """Ticker match configuration."""

    def __init__(self, *, unprefixed_uppercase: bool = True, prefixed_lowercase: bool = True, prefixed_titlecase: bool = True):
        """Return the ticker matching configuration.

        Note that a prefixed uppercase ticker, e.g. $SPY, is always matched.

        :param unprefixed_uppercase: Match SPY.
        :param prefixed_lowercase: Match $spy
        :param prefixed_titlecase: Match $Spy.
        """
        self.unprefixed_uppercase = unprefixed_uppercase
        self.prefixed_lowercase = prefixed_lowercase
        self.prefixed_titlecase = prefixed_titlecase


class TickerExtractor:
    """Ticker extractor."""

    def __init__(self, *, deduplicate: bool = True, match_config: Optional[TickerMatchConfig] = None) -> None:
        """Return the ticker extractor.

        :param deduplicate: Deduplicate the results.
        :param match_config: Optional match configuration.
        """
        self.deduplicate = deduplicate
        self.match_config = match_config or TickerMatchConfig()

    @functools.cached_property
    def pattern(self) -> Pattern:
        """Return the regular expression pattern to find possible tickers.

        This does not use the blacklist.
        """
        match_config = self.match_config
        pattern_format = r"\b{pattern}\b"
        pos_prefix, neg_prefix = r"(?<=\$)", r"(?<!\$)"
        patterns = [pos_prefix + r"[A-Z]{1,6}"]  # Match prefixed uppercase.

        if match_config.unprefixed_uppercase:
            patterns.append(neg_prefix + r"[A-Z]{2,6}")
        if match_config.prefixed_lowercase:
            patterns.append(pos_prefix + r"[a-z]{1,6}")
        if match_config.prefixed_titlecase:
            patterns.append(pos_prefix + r"[A-Z]{1}[a-z]{2,5}")

        patterns = [pattern_format.format(pattern=pattern) for pattern in patterns]
        pattern = re.compile("|".join(patterns), flags=re.ASCII)
        return pattern

    def extract(self, text: str) -> List[str]:
        """Return possible tickers extracted from the given text."""
        blacklist: Set[str] = config.BLACKLIST or set()
        matches = [match.upper() for match in self.pattern.findall(text)]
        if self.deduplicate:
            matches = list(dict.fromkeys(matches))
        matches = [match for match in matches if match not in blacklist]
        return matches



"""Test blacklist."""
# pylint: disable=missing-class-docstring,missing-function-docstring
import unittest

import reticker


class TestBlacklist(unittest.TestCase):
    def test_blacklist(self):
        blacklist = reticker.BLACKLIST
        self.assertIsInstance(blacklist, set)
        for entry in blacklist:
            self.assertIsInstance(entry, str)
            self.assertTrue(entry)
            self.assertEqual(entry.strip().upper(), entry)



"""Package installation setup."""
import os
import re
from pathlib import Path
from typing import Match, cast

from setuptools import find_packages, setup

_DIR = Path(__file__).parent
_PACKAGE_NAME = "reticker"

setup(
    name=_PACKAGE_NAME,
    author="Ouroboros Chrysopoeia",
    author_email="impredicative@users.noreply.github.com",
    version=cast(Match, re.fullmatch(r"refs/tags/v?(?P<ver>\S+)", os.environ["GITHUB_REF"]))["ver"],  # Ex: GITHUB_REF="refs/tags/1.2.3"; version="1.2.3"
    description="Use a regular expression to extract US style stock tickers from text",
    keywords="regex regexp stock text ticker",
    long_description=(_DIR / "README.md").read_text().strip(),
    long_description_content_type="text/markdown",
    url="https://github.com/impredicative/reticker/",
    packages=find_packages(exclude=["scripts"]),
    package_data={_PACKAGE_NAME: ["config/blacklist/*.txt"]},
    python_requires=">=3.8",
    classifiers=[  # https://pypi.org/classifiers/
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Topic :: Office/Business :: Financial :: Investment",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Text Processing :: Filters",
    ],
)
